﻿//
// New Script
//

const QUBES_VERSION = 'v8';

procedure ExtractZip(zip : TZipReader ; path : String);
begin
   if (path <> '') and not StrEndsWith(path, '\') then path += '\';
   for var idx := 0 to zip.Count - 1 do
      zip.Unzip(idx, path + zip.Names[idx]);
end;

var localWmi : ComVariant;

function Wmi : ComVariant;
begin
   if VarIsEmpty(localWmi) then
      localWmi := CreateOleObject('WbemScripting.SWbemLocator').ConnectServer('localhost', 'root\CIMV2', '', '');
   Result := localWmi
end;

type TService = record
   ServiceName : String;
   ExePath : String;
   ExeName : String;
   StartMode : String;
   State : String;
end;

function InstalledQubesServices : array of TService;
begin
   var objSet := Wmi.ExecQuery("SELECT * FROM Win32_Service");
   for var obj in objSet do begin
      var tmp : TService;
      tmp.ExeName := ExtractFileName(obj.PathName).ToLower;
      if tmp.ExeName not in ['qubesguardian.exe',
                             'qubesexpress.exe',
                             'qubespeon.exe',
                             'qubesibbackup.exe',
                             'p2pproxy.exe',
                             'qubesrouter.exe',
                             'qubesindexer.exe',
                             'qubesauth.exe'] then continue;
      tmp.ServiceName := obj.Name;
      tmp.ExePath := ExtractFilePath(obj.PathName);
      tmp.StartMode := obj.StartMode;
      tmp.State := obj.State;
      Result.Add(tmp);
   end;
end;

function ReadVersionInZip(zip : TZipReader) : String;
begin
   var lines := TStrings.Create;
   lines.Text := zip.Data[zip.IndexOf('version.txt')];
   if (lines[0] = 'Quality-Based Execution System') and StrBeginsWith(lines[3], 'Qubes ' + QUBES_VERSION) then
      Result := lines[3];
end;

function SelectFolder(const prompt : String) : String;
begin
   var shell := CreateOleObject('Shell.Application');
   var folder := shell.BrowseForFolder( 0, prompt, 3, '');
   if folder = nil then exit;
   Result := folder.Self.Path;
   if not Result.EndsWith('\') then Result += '\';
end;

procedure AppendFolderIfNotEmpty(path : String ; folders : array of String);
begin
   if (EnumerateDir(path, '*.*', True).Count > 0) then
      folders.Add(path);
end;

function ConfirmInstallationFolders(existingFolders : array of String) : Boolean;
begin
   Result := (existingFolders.Count = 0)
      or QuestionDlg('Following installation folders already exist :'#13#10
+ existingFolders.Join(#13#10) + #13#10#13#10
+ #'Folders content will be kept as it is, except binary files.
In order to have a fresh installation, we recommend that you rename existing folders
before installing then delete them if not useful anymore.
Do you want to continue the installation in this folders ?');
end;

//------------------------------------------------------------------------------

var zip : TZipReader;
var rootFolder := '';
var existingFolders : array of String;
var installedServices : array of TService;
var guardianIsInstalled : Boolean;

procedure CopyVBSUtilities;
begin
   SaveStringToFile(rootFolder + 'Bin\RegisterNewQubesServices.vbs', ResourceAsString('RegisterNewQubesServices.vbs'));
   SaveStringToFile(rootFolder + 'Bin\UnregisterAllQubesServices.vbs', ResourceAsString('UnregisterAllQubesServices.vbs'));
end;

procedure InitializeSDBAliasesIni;
begin
   if not FileExists(rootFolder + 'Bin\SDBAliases.ini') then
      SaveStringToTextFile(rootFolder + 'Bin\SDBAliases.ini', #'
[Qubes]
Middleware=P2PProxy
P2PServer=localhost
DBParams=Qubes

[QubesArchive]
Middleware=P2PProxy
P2PServer=localhost
DBParams=QubesArchive
');
end;

function DefaultConfigurationFileContent(name : String) : String;
begin
   Result := (if zip.IndexOf(name) >= 0 then zip.Data[zip.IndexOf(name)])
          ?? ResourceAsString(name);
end;

procedure InstallQubesGuardian(binFolderName : String = 'QubesGuardian');
begin
   var path := rootFolder + 'Bin\' + binFolderName + '\';
   ForceDirectories(path);
   ForceDirectories(rootFolder + 'Log\');
   ForceDirectories(rootFolder + 'Tmp\');

   zip.Unzip(zip.IndexOf('QubesGuardian.exe'), path + 'QubesGuardian.exe');
   zip.Unzip(zip.IndexOf('ThreadDiagnostic.exe'), path + 'ThreadDiagnostic.exe');

   var iniData := DefaultConfigurationFileContent('QubesGuardian.ini');
   SaveStringToTextFile(path + 'QubesGuardian.ini.def', iniData);
   if not FileExists(path + 'QubesGuardian.ini') then begin
      SaveStringToTextFile(path + 'QubesGuardian.ini', iniData);
      var ini := TIniFile.Create(path + 'QubesRouter.ini');
      ini.WriteString('Error.log', 'Path', rootFolder + 'Log\');
   end;

   CopyVBSUtilities;
end;

procedure InstallQubesRouter(binFolderName : String = 'QubesRouter');
begin
   var path := rootFolder + 'Bin\' + binFolderName + '\';
   ForceDirectories(path);
   ForceDirectories(rootFolder + 'Log\');
   ForceDirectories(rootFolder + 'Tmp\');

   zip.Unzip(zip.IndexOf('QubesRouter.exe'), path + 'QubesRouter.exe');

   var iniData := DefaultConfigurationFileContent('QubesRouter.ini');
   SaveStringToTextFile(path + 'QubesRouter.ini.def', iniData);
   if not FileExists(path + 'QubesRouter.ini') then begin
      SaveStringToTextFile(path + 'QubesRouter.ini', iniData);
      var ini := TIniFile.Create(path + 'QubesRouter.ini');
      ini.WriteString('Error.log', 'Path', rootFolder + 'Log\');
      ini.WriteString('Temp', 'Path', rootFolder + 'Tmp\');
   end;
   
   InitializeSDBAliasesIni;
end;

procedure InstallQubesAuth(binFolderName : String = 'QubesAuth');
begin
   var path := rootFolder + 'Bin\' + binFolderName + '\';
   ForceDirectories(path);
   ForceDirectories(rootFolder + 'Log\');
   ForceDirectories(rootFolder + 'Tmp\');
   ForceDirectories(rootFolder + 'Auth\');

   zip.Unzip(zip.IndexOf('QubesAuth.exe'), path + 'QubesAuth.exe');

   var iniData := DefaultConfigurationFileContent('QubesAuth.ini');
   SaveStringToTextFile(path + 'QubesAuth.ini.def', iniData);
   if not FileExists(path + 'QubesAuth.ini') then begin
      SaveStringToTextFile(path + 'QubesAuth.ini', iniData);
      var ini := TIniFile.Create(path + 'QubesAuth.ini');
      ini.WriteString('Error.log', 'Path', rootFolder + 'Log\');
      ini.WriteString('Temp', 'Path', rootFolder + 'Tmp\');
      ini.WriteString('Auth', 'DBPath', rootFolder + 'Auth\');
   end;

   if not FileExists(path + 'SDBAliases.ini') then
      SaveStringToTextFile(path + 'SDBAliases.ini', #'
[Qubes]
Middleware=

[QubesArchive]
Middleware=
');
end;

procedure InstallP2PProxy(binFolderName : String = 'P2PProxy');
begin
   var path := rootFolder + 'Bin\' + binFolderName + '\';
   ForceDirectories(path);
   ForceDirectories(rootFolder + 'Log\');
   ForceDirectories(rootFolder + 'Tmp\');
   ForceDirectories(rootFolder + 'Data\');
   ForceDirectories(rootFolder + 'FastSync\');

   zip.Unzip(zip.IndexOf('P2PProxy.exe'), path + 'P2PProxy.exe');
   zip.Unzip(zip.IndexOf('QuickAliasEditor.exe'), path + 'QuickAliasEditor.exe');

   SaveStringToFile(path + 'QuickAliasEditor.srv', '');

   var iniData := DefaultConfigurationFileContent('P2PProxy.ini');
   SaveStringToTextFile(path + 'P2PProxy.ini.def', iniData);
   if not FileExists(path + 'P2PProxy.ini') then begin
      SaveStringToTextFile(path + 'P2PProxy.ini', iniData);
      var ini := TIniFile.Create(path + 'P2PProxy.ini');
      ini.WriteString('Error.log', 'Path', rootFolder + 'Log\');
      ini.WriteString('Temp', 'Path', rootFolder + 'Tmp\');
      ini.WriteString('QueryLog', 'Name', rootFolder + 'FastSync\db.sql3');
   end;

   iniData := DefaultConfigurationFileContent('ErrorLog.ini');
   SaveStringToTextFile(path + 'ErrorLog.ini.def', iniData);
   if not FileExists(path + 'ErrorLog.ini') then begin
      SaveStringToTextFile(path + 'ErrorLog.ini', iniData);
      var ini := TIniFile.Create(path + 'ErrorLog.ini');
      ini.WriteString('DataBase', 'Location', rootFolder + 'Log\ERRORLOG.DB');
   end;

   if not FileExists(path + 'SDBAliases.ini') then
      SaveStringToTextFile(path + 'SDBAliases.ini', #'
[Qubes]
Middleware=
QueryLogDB=

[QubesArchive]
Middleware=
QueryLogDB=
');
end;

procedure InstallQubesIBBackup(binFolderName : String = 'QubesIBBackup');
begin
   var path := rootFolder + 'Bin\' + binFolderName + '\';
   ForceDirectories(path);
   ForceDirectories(rootFolder + 'Log\');
   ForceDirectories(rootFolder + 'Tmp\');
   ForceDirectories(rootFolder + 'Backup\');

   zip.Unzip(zip.IndexOf('QubesIBBackup.exe'), path + 'QubesIBBackup.exe');

   var iniData := DefaultConfigurationFileContent('QubesIBBackup.ini');
   SaveStringToTextFile(path + 'QubesIBBackup.ini.def', iniData);
   if not FileExists(path + 'QubesIBBackup.ini') then begin
      SaveStringToTextFile(path + 'QubesIBBackup.ini', iniData);
      var ini := TIniFile.Create(path + 'QubesIBBackup.ini');
      ini.WriteString('Error.log', 'Path', rootFolder + 'Log\');
      ini.WriteString('PARAMS', 'BACKUPFOLDER', rootFolder + 'Backup\');
   end;
   if not FileExists(path + 'SDBAliases.ini') then
      SaveStringToTextFile(path + 'SDBAliases.ini', #'
[Qubes]
Middleware=

[QubesArchive]
Middleware=
');
end;

procedure InstallQubesPeon(binFolderName : String = 'QubesPeon');
begin
   var path := rootFolder + 'Bin\' + binFolderName + '\';
   ForceDirectories(path);
   ForceDirectories(rootFolder + 'Log\');
   ForceDirectories(rootFolder + 'Tmp\');
   ForceDirectories(rootFolder + 'FastSync\');

   zip.Unzip(zip.IndexOf('QubesPeon.exe'), path + 'QubesPeon.exe');

   var iniData := DefaultConfigurationFileContent('QubesPeon.ini');
   SaveStringToTextFile(path + 'QubesPeon.ini.def', iniData);
   if not FileExists(path + 'QubesPeon.ini') then begin
      SaveStringToTextFile(path + 'QubesPeon.ini', iniData);
      var ini := TIniFile.Create(path + 'QubesPeon.ini');
      ini.WriteString('Error.log', 'Path', rootFolder + 'Log\');
      ini.WriteString('Temp', 'Path', rootFolder + 'Tmp\');
      ini.WriteString('QueryLog', 'Name', rootFolder + 'FastSync\db.sql3');
      ini.WriteString('XXL', 'Delay', '0');
   end;
   InitializeSDBAliasesIni;
end;

procedure InstallQubesIndexer(binFolderName : String = 'QubesIndexer');
begin
   var path := rootFolder + 'Bin\' + binFolderName + '\';
   ForceDirectories(path);
   ForceDirectories(rootFolder + 'Log\');
   ForceDirectories(rootFolder + 'Tmp\');
   ForceDirectories(rootFolder + 'Idx\Docs\');
   ForceDirectories(rootFolder + 'Idx\Procs\');
   ForceDirectories(rootFolder + 'Idx\Mediums\');

   zip.Unzip(zip.IndexOf('QubesIndexer.exe'), path + 'QubesIndexer.exe');
   zip.Unzip(zip.IndexOf('ifilterconv.exe'), path + 'ifilterconv.exe');
   zip.Unzip(zip.IndexOf('ifilterconv64.exe'), path + 'ifilterconv64.exe');
   zip.Unzip(zip.IndexOf('pdftotext.exe'), path + 'pdftotext.exe');

   var iniData := DefaultConfigurationFileContent('QubesIndexer.ini');
   SaveStringToTextFile(path + 'QubesIndexer.ini.def', iniData);
   if not FileExists(path + 'QubesIndexer.ini') then begin
      SaveStringToTextFile(path + 'QubesIndexer.ini', iniData);
      var ini := TIniFile.Create(path + 'QubesIndexer.ini');
      ini.WriteString('Error.log', 'Path', rootFolder + 'Log\');
      ini.WriteString('Temp', 'Path', rootFolder + 'Tmp\');
      ini.WriteString('Qubes', 'DocsIndex', rootFolder + 'Idx\Docs\');
      ini.WriteString('Qubes', 'ProcsIndex', rootFolder + 'Idx\Procs\');
      ini.WriteString('Qubes', 'MediumsIndex', rootFolder + 'Idx\Mediums\');
   end;
   
   InitializeSDBAliasesIni;
end;

procedure InstallQubesExpress(binFolderName : String = 'QubesExpress');
begin
   var path := rootFolder + 'Bin\' + binFolderName + '\';
   ForceDirectories(path);
   ForceDirectories(rootFolder + 'Log\');
   ForceDirectories(rootFolder + 'Tmp\');
   ForceDirectories(rootFolder + 'Bin\GS\');
   ForceDirectories(rootFolder + 'Bin\PDFTK\');
   ForceDirectories(rootFolder + 'Bin\msxml\');
   ForceDirectories(rootFolder + 'Bin\wkhtmltopdf\');

   zip.Unzip(zip.IndexOf('QubesExpress.exe'), path + 'QubesExpress.exe');
   zip.Unzip(zip.IndexOf('MainCfgFileEditor.exe'), path + 'MainCfgFileEditor.exe');

   // Install needed companion files
   if binFolderName = 'QubesExpress' then begin
      zip.Unzip(zip.IndexOf('msxml4.dll'), rootFolder + 'Bin\msxml\msxml4.dll');
      zip.Unzip(zip.IndexOf('msxml4a.dll'), rootFolder + 'Bin\msxml\msxml4a.dll');
      zip.Unzip(zip.IndexOf('msxml4r.dll'), rootFolder + 'Bin\msxml\msxml4r.dll');
      zip.Unzip(zip.IndexOf('msxml6.dll'), rootFolder + 'Bin\msxml\msxml6.dll');
      zip.Unzip(zip.IndexOf('msxml6r.dll'), rootFolder + 'Bin\msxml\msxml6r.dll');
      zip.Unzip(zip.IndexOf('pdftk.exe'), rootFolder + 'Bin\PDFTK\pdftk.exe');
      zip.Unzip(zip.IndexOf('wkhtmltoimage.exe'), rootFolder + 'Bin\wkhtmltopdf\wkhtmltoimage.exe');
      zip.Unzip(zip.IndexOf('wkhtmltox.dll'), rootFolder + 'Bin\wkhtmltopdf\wkhtmltox.dll');
      zip.Unzip(zip.IndexOf('wkhtmltopdf.exe'), rootFolder + 'Bin\wkhtmltopdf\wkhtmltopdf.exe');
      ExtractZip(TZipReader.FromData(zip.Data[zip.IndexOf('ghostscript.zip')]), rootFolder + '\Bin\GS\');
   end;

   var iniData := DefaultConfigurationFileContent('QubesExpress.ini');
   SaveStringToTextFile(path + 'QubesExpress.ini.def', iniData);
   if not FileExists(path + 'QubesExpress.ini') then begin
      SaveStringToTextFile(path + 'QubesExpress.ini', iniData);
      var ini := TIniFile.Create(path + 'QubesExpress.ini');
      ini.WriteString('Error.log', 'Path', rootFolder + 'Log\');
      ini.WriteString('Temp', 'Path', rootFolder + 'Tmp\');
   end;

   iniData := DefaultConfigurationFileContent('server.cfg');
   SaveStringToTextFile(path + 'server.cfg.def', iniData);
   if not FileExists(path + 'server.cfg') then
      SaveStringToTextFile(path + 'server.cfg', iniData);

   InitializeSDBAliasesIni;
end;

procedure InstallQubes(binFolderName : String = 'Qubes');
begin
   var path := rootFolder + 'Bin\' + binFolderName + '\';
   ForceDirectories(path);
   ForceDirectories(rootFolder + 'Log\');
   ForceDirectories(rootFolder + 'Tmp\');
   ForceDirectories(path + 'Models\XfmTemplates\');

   zip.Unzip(zip.IndexOf('Qubes.exe'), path + 'Qubes.exe');
   zip.Unzip(zip.IndexOf('Qubes.qsh'), path + 'Qubes.qsh');
   zip.Unzip(zip.IndexOf('Qubes.nch'), path + 'Qubes.nch');
   zip.Unzip(zip.IndexOf('xfm.qsh'), path + 'xfm.qsh');
   // Chromium
   ExtractZip(TZipReader.FromData(zip.Data[zip.IndexOf('chromium3.zip')]), rootFolder + 'Bin\');
   // Models
   zip.Unzip(zip.IndexOf('bleu et blanc.xfmt'), path + 'Models\XfmTemplates\bleu et blanc.xfmt');
   zip.Unzip(zip.IndexOf('ColorPicker.xfmlt'), path + 'Models\XfmTemplates\ColorPicker.xfmlt');
   zip.Unzip(zip.IndexOf('valider suspendre.xfmlt'), path + 'Models\XfmTemplates\valider suspendre.xfmlt');
   ExtractZip(TZipReader.FromData(zip.Data[zip.IndexOf('modele.zip')]), path + 'Models\');

   var iniData := DefaultConfigurationFileContent('Qubes.ini');
   SaveStringToTextFile(path + 'Qubes.ini.def', iniData);
   if not FileExists(path + 'Qubes.ini') then begin
      SaveStringToTextFile(path + 'Qubes.ini', iniData);
      var ini := TIniFile.Create(path + 'Qubes.ini');
      ini.WriteString('Error.log', 'Path', rootFolder + 'Log\');
      ini.WriteString('Temp', 'Path', rootFolder + 'Tmp\');
   end;

   InitializeSDBAliasesIni;
end;

procedure InstallTools(binFolderName : String = 'Tools');
begin
   var path := rootFolder + 'Bin\' + binFolderName + '\';
   ForceDirectories(path);

   zip.Unzip(zip.IndexOf('ConsoleSQL.exe'), path + 'ConsoleSQL.exe');
   zip.Unzip(zip.IndexOf('MiniDelphi.exe'), path + 'MiniDelphi.exe');
   zip.Unzip(zip.IndexOf('MiniQubes.exe'), path + 'MiniQubes.exe');
   zip.Unzip(zip.IndexOf('ErrorLogPathCfg.exe'), path + 'ErrorLogPathCfg.exe');
   zip.Unzip(zip.IndexOf('LDAPTest.exe'), path + 'LDAPTest.exe');
   zip.Unzip(zip.IndexOf('SMTPTester.exe'), path + 'SMTPTester.exe');
   zip.Unzip(zip.IndexOf('QueryLogBackup.exe'), path + 'QueryLogBackup.exe');
   zip.Unzip(zip.IndexOf('QuBESDatabasesBackup.exe'), path + 'QuBESDatabasesBackup.exe');
   zip.Unzip(zip.IndexOf('register.vnd.sun.star.webdav.bat'), path + 'register.vnd.sun.star.webdav.bat');
end;

//------------------------------------------------------------------------------

function InputFolderName(const caption : String) : String;
begin
   function CheckFolderInUse(folderPath : String ; interactive : Boolean = True) : Boolean;
   begin
      for var service in installedServices do
         if (service.ExePath.ToLower = folderPath.ToLower) then begin
            if interactive then
               ErrorDlg('Path "' + folderPath + '" is already used for an installed service.');
            exit(True);
         end;
   end;

   while True do begin
      Result := InputBox('Folder name', caption, '');
      if (Result = '') then
         exit
      else if (Pos('\', Result) > 0) then
         ErrorDlg('Invalid folder name "' + Result + '"')
      else if not CheckFolderInUse(rootFolder + 'Bin\' + Result) then
         break;
   end;
end;

procedure Action_InstallQubesIBBackup;
begin
   var serviceIsInstalled := False;
   for var service in installedServices do
      if service.ExeName = 'qubesibbackup.exe' then begin
         serviceIsInstalled := True;
         break;
      end;

   var folderName := 'QubesIBBackup';
   if serviceIsInstalled then begin
      ErrorDlg(#'QubesIBBackup is already installed.
If you want to replace current installation, you must uninstall it first.');
      exit;
   end;

   // Confirmation if installation folders already exist
   if not guardianIsInstalled then
      AppendFolderIfNotEmpty(rootFolder + 'Bin\QubesGuardian', existingFolders);
   AppendFolderIfNotEmpty(rootFolder + 'Bin\' + folderName, existingFolders);
   if not ConfirmInstallationFolders(existingFolders) then exit;

   // Install
   var pw := TProgressWait.Create('Installation', 'Installation', False);
   pw.Start;
   try
      if not guardianIsInstalled then
         InstallQubesGuardian;
      InstallQubesIBBackup(folderName);
   finally
      pw.Stop;
   end;

   InformationDlg(#'QubesIBBackup is successfully installed on this system.
Change settings in configuration files then execute "\Bin\RegisterNewQubesServices.vbs"');
end;

procedure Action_InstallQubesRouter;
begin
   var serviceIsInstalled := False;
   for var service in installedServices do
      if service.ExeName = 'qubesrouter.exe' then begin
         serviceIsInstalled := True;
         break;
      end;

   var folderName := 'QubesRouter';
   if serviceIsInstalled then begin
      if not QuestionDlg(#'QubesRouter is already registered as a service.
If you want to replace current installation, you must uninstall it first.
If you want a side-by-side installation, be careful with the settings in "QubesRouter.ini".
Do you want a side-by-side installation ?') then exit;
      folderName := InputFolderName('Side-by-side QubesRouter folder name');
      if (folderName = '') then exit;
   end;

   // Confirmation if installation folders already exist
   if not guardianIsInstalled then
      AppendFolderIfNotEmpty(rootFolder + 'Bin\QubesGuardian', existingFolders);
   AppendFolderIfNotEmpty(rootFolder + 'Bin\' + folderName, existingFolders);
   if not ConfirmInstallationFolders(existingFolders) then exit;

   // Install
   var pw := TProgressWait.Create('Installation', 'Installation', False);
   pw.Start;
   try
      if not guardianIsInstalled then
         InstallQubesGuardian;
      InstallQubesRouter(folderName);
   finally
      pw.Stop;
   end;

   InformationDlg(#'QubesRouter is successfully installed on this system.
Change settings in configuration files then execute "\Bin\RegisterNewQubesServices.vbs"');
end;

procedure Action_InstallP2PProxy;
begin
   var serviceIsInstalled := False;
   for var service in installedServices do
      if service.ExeName = 'p2pproxy.exe' then begin
         serviceIsInstalled := True;
         break;
      end;

   var folderName := 'P2PProxy';
   if serviceIsInstalled then begin
      if not QuestionDlg(#'SDB P2PProxy is already registered as a service.
If you want to replace current installation, you must uninstall it first.
If you want a side-by-side installation, be careful with the settings in "P2PProxy.ini".
Do you want a side-by-side installation ?') then exit;
      folderName := InputFolderName('Side-by-side P2PProxy folder name');
      if folderName = '' then exit;
   end;

   // Confirmation if installation folders already exist
   if not guardianIsInstalled then
      AppendFolderIfNotEmpty(rootFolder + 'Bin\QubesGuardian', existingFolders);
   AppendFolderIfNotEmpty(rootFolder + 'Bin\' + folderName, existingFolders);
   if not ConfirmInstallationFolders(existingFolders) then exit;

   // Install
   var pw := TProgressWait.Create('Installation', 'Installation', False);
   pw.Start;
   try
      if not guardianIsInstalled then
         InstallQubesGuardian;
      InstallP2PProxy(folderName);
   finally
      pw.Stop;
   end;

   InformationDlg(#'SDB P2PProxy is successfully installed on this system.
Change settings in configuration files then execute "\Bin\RegisterNewQubesServices.vbs"');
end;

procedure Action_InstallQubesExpress;
begin
   var serviceIsInstalled := False;
   for var service in installedServices do
      if service.ExeName = 'qubesexpress.exe' then begin
         serviceIsInstalled := True;
         break;
      end;

   var folderName := 'QubesExpress';
   if serviceIsInstalled then begin
      if not QuestionDlg(#'QubesExpress is already registered as a service.
If you want to replace current installation, you must uninstall it first.
If you want a side-by-side installation, be careful with the settings in "server.cfg".
Do you want a side-by-side installation ?') then exit;
      folderName := InputFolderName('Side-by-side QubesExpress folder name');
      if folderName = '' then exit;
   end;

   // Confirmation if installation folders already exist
   if not guardianIsInstalled then
      AppendFolderIfNotEmpty(rootFolder + 'Bin\QubesGuardian', existingFolders);
   AppendFolderIfNotEmpty(rootFolder + 'Bin\' + folderName, existingFolders);
   if not ConfirmInstallationFolders(existingFolders) then exit;

   // Install
   var pw := TProgressWait.Create('Installation', 'Installation', False);
   pw.Start;
   try
      if not guardianIsInstalled then
         InstallQubesGuardian;
      InstallQubesExpress(folderName);
   finally
      pw.Stop;
   end;

   InformationDlg(#'QubesExpress is successfully installed on this system.
Change settings in configuration files then execute "\Bin\RegisterNewQubesServices.vbs"');
end;

procedure Action_InstallStandardServices;
begin
   if (installedServices.Count > 0) then begin
      ErrorDlg(#'Qubes services are already installed.
If you want to replace current installation, you must uninstall them first.');
      exit;
   end;
   
   // Confirmation if installation folders already exist
   for var folder in ['QubesGuardian', 'P2PProxy', 'QubesPeon', 'QubesIndexer', 'QubesExpress', 'Qubes'] do
      AppendFolderIfNotEmpty(rootFolder + 'Bin\' + folder, existingFolders);
   if not ConfirmInstallationFolders(existingFolders) then exit;

   // Install
   var pw := TProgressWait.Create('Installation', 'Installation', False);
   pw.Start;
   try
      if not guardianIsInstalled then
         InstallQubesGuardian;
      InstallP2PProxy;
      InstallQubesAuth;
      InstallQubesPeon;
      InstallQubesIndexer;
      InstallQubesExpress;
      InstallQubes;
      InstallTools;
   finally
      pw.Stop;
   end;

   InformationDlg(#'Services are successfully installed on this system.
Change settings in configuration files then execute "\Bin\RegisterNewQubesServices.vbs"');
end;

//------------------------------------------------------------------------------

// zip file selection
var zipFileName := CmdLineParamStr(1);
if (zipFileName = '') then
   zipFileName := OpenDialog('', '', 'zip|*.zip');

var version := '';
if (zipFileName <> '') and FileExists(zipFileName) then try
   zip := TZipReader.Create(zipFileName);
   version := ReadVersionInZip(zip);
except
end;

// check zip file
if version = '' then begin
   var err := '';
   if zipFileName <> '' then
      err := 'Invalid zip file "' + zipFileName + '" for Qubes ' + QUBES_VERSION + #13#10;
   err += 'Download the desired Qubes installation zip file on "https://support.creative-it.net" then drag & drop it on QubesInstaller.exe';
   ErrorDlg(err);
   exit;
end;

// get installed services
installedServices := InstalledQubesServices;
for var service in installedServices do begin
   rootFolder := service.ExePath;
   if service.ExeName = 'qubesguardian.exe' then begin
      guardianIsInstalled := True;
      break;
   end;
end;

// define installation folder if there is no installed service
if (rootFolder = '') then begin
   rootFolder := SelectFolder('Select installation folder (ex : "D:\Qubes\")');
   if (rootFolder = '') or rootFolder.StartsWith('\\') then exit;
   if rootFolder.EndsWith(':\') then
      rootFolder += 'Qubes\';
end else
   rootFolder := StrBeforeLast(StrBeforeLast(rootFolder.DeleteRight(1), '\'), '\') + '\';

// Main form
while True do begin
   existingFolders.Clear;
   var settings := TStrings.Create;
   settings.CommaText := HTMLBox(#'<html>
<head><style>
body { font-family: Arial; text-align: center; font-size: 14px; }
td, input { vertical-align: middle; height: 40px; padding: 5px 10px; }
</style></head><body>
<img src="LogoCIT.png">
<h1 style="color: #0086FF;">Installation of ' + HTMLTextEncoder.Encode(version) + #'</h1>
<table>
<tr><td colspan="2"><u>Installation folder</u>: <b>' + rootFolder.ToHtml + #'</b></td></tr>
<tr><td><input type="submit" name="Full" value="-->"></td><td>Install standard Qubes services</td></tr>
<tr><td><input type="submit" name="IbBackup" value="-->"></td><td>Install QubesIbBackup<br>(only useful with Firebird databases)</td></tr>
<tr><td><input type="submit" name="Router" value="-->"></td><td>Install QubesRouter<br>(stand-alone or side-by-side)</td></tr>
<tr><td><input type="submit" name="P2PProxy" value="-->"></td><td>Install P2PProxy<br>(stand-alone or side-by-side)</td></tr>
<tr><td><input type="submit" name="Express" value="-->"></td><td>Install QubesExpress<br>(stand-alone or side-by-side)</td></tr>
</table>
</body></html>', 400, 520);
   if settings.Values['Router'] <> '' then
      Action_InstallQubesRouter
   else if settings.Values['IbBackup'] <> '' then
      Action_InstallQubesIBBackup
   else if settings.Values['P2PProxy'] <> '' then
      Action_InstallP2PProxy
   else if settings.Values['Express'] <> '' then
      Action_InstallQubesExpress
   else if settings.Values['Full'] <> '' then
      Action_InstallStandardServices
   else exit;
end;
